#include <stdio.h>
#include <stdlib.h>
#include "toy.h"

struct symbol_info {
   char *name;
   void *place;
   int type;
};

struct type_info {
   int var_type;
   int base;
   int mod;
   int rettype;
};

char *str_dup(char *source);

struct symbol_info sym_table[MAX_SYMBOLS];
struct type_info types[MAX_TYPES];
int display[MAX_LEVELS+1];
int current_level = -1;
int num_of_types = 1;


void init_prog(void)
{
   /* one fundamental data type: integer */
   types[0].var_type = TYPE_REGULAR;
   types[0].base = 0;
   types[0].mod = 0;

   /* empty arg list */
   types[1].var_type = TYPE_REGULAR;
   types[1].base = 0;
   types[1].mod = 0;

   /* make 'get():integer' */
   add_symbol("get", make_function(TYPE_EMPTYLIST, TYPE_INTEGER));
   
   /* make 'put(integer):integer' */
   add_symbol("put", make_function(TYPE_INTEGER, TYPE_INTEGER));
}


int make_type(int base, int modifier)
{
   struct type_info dummy;
   int returnval;

   if (num_of_types >= MAX_TYPES) {
      yyerror("internal: no space left to add new type");
      return -1;
   }

   dummy.var_type = types[base].var_type;
   dummy.base = base;
   dummy.mod = modifier;
   returnval = num_of_types;

   types[num_of_types++] = dummy;

#ifdef DEBUG
   fprintf(stderr, "Making type %d: %d, %d\n", returnval, base, modifier);
#endif

   return returnval;
}


int get_type(int base, int modifier)
{
   int i;

   for (i = 1; i < num_of_types; i++)
      if ((types[i].var_type == TYPE_REGULAR) && (types[i].base == base) &&
	  (types[i].mod == modifier))
	 return i;

   return (make_type(base, modifier));
}


int make_function(int base, int rettype)
{
   struct type_info dummy;
   int returnval;

   if (num_of_types >= MAX_TYPES) {
      yyerror("internal: no space left to add new type");
      return -1;
   }

   dummy.var_type = TYPE_FUNCTION;
   dummy.base = base;
   dummy.rettype = rettype;
   returnval = num_of_types;

   types[num_of_types++] = dummy;

#ifdef DEBUG
   fprintf(stderr, "Making function %d: %d, %d\n", returnval, base, rettype);
#endif

   return returnval;
}


int get_function(int base, int rettype)
{
   int i;

   for (i = 1; i < num_of_types; i++)
      if ((types[i].var_type == TYPE_FUNCTION) && (types[i].base == base) &&
	  (types[i].rettype == rettype))
	 return i;

   return (make_function(base, rettype));
}



int is_function(int type)
{
   if (type < 0 || type >= num_of_types) {
      yyerror("internal: inconsistency in internal type tables");
      return 0;
   }

   return (types[type].var_type == TYPE_FUNCTION);
}


int arg_list(int type)
{
   if (type < 0||type >= num_of_types||types[type].var_type != TYPE_FUNCTION) {
      yyerror("internal: inconsistency in internal type tables");
      return 0;
   }

   return (types[type].base);
}


int ret_type(int type)
{
   if (type < 0||type >= num_of_types||types[type].var_type != TYPE_FUNCTION) {
      yyerror("internal: inconsistency in internal type tables");
      return 0;
   }

   return (types[type].rettype);
}


int deref_type(int type, int mod)
{
   if (type < 0 || type >= num_of_types) {
      yyerror("internal: inconsistency in internal type tables");
      return -1;
   }

   if (types[type].mod != mod)
      return -1;
   else
      return types[type].base;
}


void block_start(void)
{
   char buf[128];
#ifdef DEBUG
   fprintf(stderr, "Starting block.\n");
#endif

   if (current_level >= MAX_LEVELS) {
      sprintf(buf, "too many lexical levels (max %d)", MAX_LEVELS);
      yyerror(buf);
      return;
   }

   if (current_level < 0) {
      current_level = 0;
      display[0] = 0;
   } else {
      current_level++;
      display[current_level] = display[current_level-1];
   }
}


void block_end(void)
{
#ifdef DEBUG
   fprintf(stderr, "Ending block.\n");
#endif

   current_level--;
}



void add_symbol(char *name, int type)
{
   struct symbol_info dummy;
   int i, limit;

   if (current_level < 0) {
      yyerror("internal: sym table: trying to add symbol outside of all blocks");
      return;
   }

   if (display[current_level] >= MAX_SYMBOLS) {
      yyerror("internal: no space left to add new symbol");
      return;
   }

   dummy.name  = str_dup(name);
   dummy.place = NULL; /* this must be changed later */
   dummy.type  = type;

   if (current_level == 0)
      limit = 0;
   else
      limit = display[current_level-1];

   for (i = display[current_level]-1; i >= limit; i--)
      if (!strcmp(dummy.name, sym_table[i].name)) {
	 yyerror("identifier already declared in this block");
	 return;
      }

   sym_table[display[current_level]++] = dummy;

#ifdef DEBUG
   fprintf(stderr, "Adding symbol: %s, %d\n", name, type);
#endif
}


static struct symbol_info *find_symbol(char *name)
{
   int i;

   if (current_level < 0) {
      yyerror("internal: searching for sym outside of all blocks");
      return NULL;
   }

   for (i = display[current_level] - 1; i >= 0; i--)
      if (!strcmp(name, sym_table[i].name))
	 return sym_table + i;

   return NULL;
}


int symbol_exists(char *name)
{
#ifdef DEBUG
   fprintf(stderr, "Checking symbol: %s\n", name);
#endif

   if (find_symbol(name))
      return 1;
   else
      return 0;
}


int get_symbol_type(char *name)
{
   struct symbol_info *sym;

   if ((sym = find_symbol(name)) != NULL) {
#ifdef DEBUG
      fprintf(stderr, "Type of %s is %d\n", name, sym->type);
#endif
      return sym->type;
   }
   else
      return -1;
}


/* Create a duplicate of a string */
char	*str_dup(char *source)
{
   char	*new;

   new = malloc(sizeof(char) * (strlen(source) + 1));
   return(strcpy(new, source));
}
